from abc import ABC, abstractmethod
from typing import Any, Callable, Mapping, Sequence, Tuple, TypeVar

from rx.core.typing import Observable, Observer

ReduxRootState = Mapping[str, Any]

PayloadType = TypeVar("PayloadType")

Action = Tuple[str, PayloadType]

Epic = Callable[[Observable[Action], Observable[ReduxRootState]], Observable[Action]]

StateType = TypeVar("StateType")

Reducer = Callable[[StateType, Action[PayloadType]], StateType]

ReduxFeatureModule = Tuple[str, Reducer, Epic, Sequence]


class AbstractReduxRootStore(ABC):
    """ReduxRootStore abstract base class. Untyped."""

    @abstractmethod
    def as_observable(self):
        raise NotImplementedError

    @abstractmethod
    def dispatch(self, action):
        raise NotImplementedError

    @abstractmethod
    def add_feature_module(self, module):
        raise NotImplementedError


class ReduxRootStore(AbstractReduxRootStore, Observer[Action]):
    """ReduxRootStore abstract base class."""

    @abstractmethod
    def as_observable(self) -> Observable[ReduxRootState]:
        """ Returns the state as an observable

        Returns:
            The observable version of the store
        """

        return NotImplemented

    @abstractmethod
    def dispatch(self, action: Action) -> Action:
        """ Dispatches an action. It is the only way to trigger a state change.
    
        The `reducer` function, used to create the store, will be called with the
        current state tree and the given `action`. Its return value will be
        considered the **next** state of the tree, and the change listeners will
        be notified.

        Args:
            action: A tuple representing “what changed”. The first element in the tuple
                    is considered to be the identifier of the action, the second element
                    is the action payload

        Returns:
            The same action as was passed in
        """

        return NotImplemented

    @abstractmethod
    def add_feature_module(self, module: ReduxFeatureModule) -> None:
        """ Registers a feature module with the root store

        Args:
            module: the feature model

        """

        return NotImplemented



from typing import (
    Any,
    Callable,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from rx import merge, never
from rx.core.typing import Observable

from .types import Action, Epic, PayloadType, ReduxRootState, StateType


def run_epic(
    action_: Observable[Action], state_: Observable[ReduxRootState]
) -> Callable[[Epic], Observable[ReduxRootState]]:
    """ Runs a single epic agains the given action and state observables """
    return lambda epic: epic(action_, state_)


def combine_epics(*epics: Sequence[Epic]) -> Epic:
    """ Combines a sequence of epics into one single epic by merging them """
    return lambda action_, state_: merge(*map(run_epic(action_, state_), epics))



""" Module doc """

# Version of ReduxPy package
__version__ = "0.0.1"

from .store import create_store
from .action import create_action, select_action_payload, of_type
from .reducer import handle_actions
from .feature import of_init_feature, create_feature_module, select_feature
from .epic import combine_epics
from .types import Action, Reducer, Epic, ReduxRootStore, ReduxFeatureModule



from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from .action import select_action_type
from .types import Action, PayloadType, Reducer, StateType


def default_reducer(initial_state: Optional[StateType]) -> Reducer:
    """ Creates a reducer that returns the original state or the default state. """
    return lambda state, action: state if state else initial_state


def handle_actions(
    action_map: Mapping[str, Reducer[StateType, PayloadType]],
    initial_state: Optional[StateType] = None,
) -> Reducer:
    """ Creates a new reducer from a mapping of action name to reducer for that action. """
    def_reducer = default_reducer(initial_state)
    return lambda state, action: action_map.get(
        select_action_type(action), def_reducer
    )(state, action)


def combine_reducers(reducers: Mapping[str, Reducer]) -> Reducer:
    """ Creates a new reducer from a mapping of reducers. """
    items = reducers.items()

    def combine(state: Dict[str, StateType], action: Action) -> Mapping[str, StateType]:
        """ Updates the state object from the reducer mappings. """
        result = state
        for key, value in items:
            current = state.get(key)
            updated = value(current, action)
            if not current is updated:
                if result is state:
                    result = dict(state)
                result[key] = updated
        return result

    return combine



"""Setup script for redux-py"""

from os.path import abspath, dirname, join
from setuptools import setup, find_packages

# The directory containing this file
HERE = abspath(dirname(__file__))

# The text of the README file
with open(join(HERE, "README.md")) as fid:
    README = fid.read()

# This call to setup() does all the work
setup(
    name="redux-py",
    version="0.0.1",
    description="Redux implementation for Python",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Carsten-Leue/ReduxPY",
    author="Dr. Carsten Leue",
    author_email="carsten.leue@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["rx"],
    entry_points={"console_scripts": ["reduxpy=redux.__main__:main"]},
)



from rx import Observable, pipe
from rx.operators import do_action, filter, map

from redux import (
    Epic,
    Reducer,
    ReduxFeatureModule,
    combine_epics,
    create_action,
    create_feature_module,
    handle_actions,
    of_init_feature,
    of_type,
    select_action_payload,
    select_feature,
)

INIT_FEATURE = "INIT_FEATURE"

ADD_SAMPLE_ACTION = "ADD_SAMPLE_ACTION"

add_sample_action = create_action(ADD_SAMPLE_ACTION)

sample_reducer = handle_actions(
    {ADD_SAMPLE_ACTION: lambda state, action: select_action_payload(action)}
)

add_epic: Epic = lambda action_, state_: action_.pipe(
    of_type(ADD_SAMPLE_ACTION),
    do_action(lambda action: print(f"addEpic: {action}")),
    filter(lambda x: False),
)

init_epic: Epic = lambda action_, state_: action_.pipe(
    of_init_feature(INIT_FEATURE), map(lambda x: add_sample_action("init")),
)

sample_epic: Epic = combine_epics(add_epic, init_epic)

init_feature_module: ReduxFeatureModule = create_feature_module(
    INIT_FEATURE, sample_reducer, sample_epic
)

select_init_feature_module = select_feature(INIT_FEATURE)



from typing import Any, Callable, Generic, Optional, Tuple, TypeVar, Union

from rx.core.typing import Observable
from rx.operators import filter

from .types import Action, PayloadType


def create_action(ty: str) -> Callable[[PayloadType], Action[PayloadType]]:
    """ Action creator """
    return lambda payload: (ty, payload)


def select_action_type(action: Action) -> str:
    """ selects the type from the action """
    return action[0]


def select_action_payload(action: Action[PayloadType]) -> PayloadType:
    """ selects the payload from the action """
    return action[1]


def is_type(id: str) -> Callable[[Action], bool]:
    """ Returns a function that checks if the action is of a particular type """
    return lambda action: select_action_type(action) is id


of_type: Callable[
    [str], Callable[[Observable[Action]], Observable[Action]]
] = lambda id: filter(is_type(id))



from typing import Any, Callable, Mapping, Optional, Sequence, List, Dict

from rx import merge, never
from rx.core.typing import Observable
from rx.operators import (
    distinct,
    do_action,
    filter,
    flat_map,
    map,
    scan,
    share,
    take_until,
)
from rx.subject import BehaviorSubject, Subject

from .action import create_action
from .constants import INIT_ACTION
from .epic import run_epic
from .reducer import combine_reducers
from .types import (
    Action,
    Epic,
    Reducer,
    ReduxFeatureModule,
    ReduxRootState,
    ReduxRootStore,
)

init_feature_action = create_action(INIT_ACTION)

select_id: Callable[[ReduxFeatureModule], str] = lambda module: module[0]
select_dependencies: Callable[
    [ReduxFeatureModule], Sequence[ReduxFeatureModule]
] = lambda module: module[3]
select_reducer: Callable[[ReduxFeatureModule], Reducer] = lambda module: module[1]
select_epic: Callable[[ReduxFeatureModule], Epic] = lambda module: module[2]

has_reducer: Callable[[ReduxFeatureModule], bool] = lambda module: bool(
    select_reducer(module)
)

is_never = lambda x: False


identity_reducer: Reducer = lambda state, action: state


def create_store(initial_state: Optional[ReduxRootState] = {}) -> ReduxRootStore:
    """ Constructs a new store that can handle feature modules. 
    
        Args:
            initial_state: optional initial state of the store, will typically be the empty dict

        Returns:
            An implementation of the store
    """

    # current reducer
    reducer: List[Reducer] = [identity_reducer]

    def replace_reducer(new_reducer: Reducer) -> None:
        reducer[0] = new_reducer

    actions = Subject()

    actions_ = actions.pipe(
        do_action(lambda a: print("dispatching %s" % (str(a)))), share()
    )

    def _dispatch(action: Action) -> None:
        actions.on_next(action)

    state = BehaviorSubject(initial_state)

    # shutdown trigger
    done_ = Subject()

    # The set of known modules, to avoid cycles and duplicate registration
    modules: Dict[str, ReduxFeatureModule] = {}

    # Sequence of added modules
    module_subject = Subject()

    # Subscribe to the resolved modules
    module_ = module_subject.pipe(distinct(select_id), share())

    # Build the reducers
    reducer_ = module_.pipe(
        filter(has_reducer),
        scan(
            lambda all, module: {**all, select_id(module): select_reducer(module)}, {}
        ),
        map(combine_reducers),
        map(replace_reducer),
    )

    # Build the epic
    epic_ = module_.pipe(map(select_epic), filter(bool))

    # Root epic that combines all of the incoming epics
    root_epic: Epic = lambda action_, state_: epic_.pipe(
        flat_map(run_epic(action_, state_)), map(_dispatch),
    )

    # notifications about new feature states
    new_module_ = module_.pipe(
        map(select_id), map(init_feature_action), map(_dispatch),
    )

    def _add_feature_module(module: ReduxFeatureModule):
        """ Registers a new feature module """
        module_id = select_id(module)
        if not module_id in modules:
            modules[module_id] = module
            for dep in select_dependencies(module):
                _add_feature_module(dep)
            module_subject.on_next(module)

    #: all state
    internal_ = merge(root_epic(actions_, state), reducer_, new_module_).pipe(
        filter(is_never)
    )

    def _as_observable() -> Observable:
        return state

    _on_next = _dispatch
    _on_complete = lambda: done_.on_next(None)

    merge(actions_, internal_).pipe(
        map(lambda action: reducer[0](state.value, action)), take_until(done_),
    ).subscribe(state)

    class ReduxRootStoreImpl(ReduxRootStore):
        """ Implementation of the ReduxRootStore. We use the class
            only as an interface and dispatch to the closure for
            all implementation
        """

        def dispatch(self, action: Action) -> Action:
            return _dispatch(action)

        def as_observable(self) -> Observable[ReduxRootState]:
            return _as_observable()

        def on_next(self, value: Action) -> None:
            _on_next(value)

        def on_completed(self) -> None:
            _on_complete()

        def on_error(self, error) -> None:
            pass

        def add_feature_module(self, module: ReduxFeatureModule) -> None:
            _add_feature_module(module)

    return ReduxRootStoreImpl()



from typing import Any, Callable, Mapping, Optional, Sequence, Tuple

from rx import pipe
from rx.core.typing import Observable
from rx.operators import do_action, filter, map, take

from .action import select_action_payload, select_action_type
from .constants import INIT_ACTION
from .types import Action, Epic, Reducer, StateType, ReduxFeatureModule

is_init_feature_action: Callable[[Action], bool] = lambda action: (
    select_action_type(action) is INIT_ACTION
)

def has_payload(payload: Any) -> Callable[[Action], bool]:
    """ Tests if the action payload matches the given payload """
    return lambda action: select_action_payload(action) is payload


def of_init_feature(id: str) -> Callable[[Observable[Action]], Observable[str]]:
    """ Operator to test for the initialization action of a feature """
    is_payload = has_payload(id)
    return pipe(
        filter(is_init_feature_action), filter(is_payload), take(1), map(lambda x: id)
    )


def create_feature_module(
    id: str,
    reducer: Reducer = None,
    epic: Epic = None,
    dependencies: Sequence[ReduxFeatureModule] = [],
) -> ReduxFeatureModule:
    """ Constructs a new feature module descriptor """
    return (id, reducer, epic, dependencies)


def select_feature(
    id: str, initial_state: StateType = None
) -> Callable[[Mapping[str, StateType]], Optional[StateType]]:
    """ Selects a feature state """
    return lambda state: state.get(id, initial_state)
